#! /usr/bin/env python
"""This module add some useful stream classes"""

import errno
import io
import logging
import os
import threading
import time

from .py26 import memoryview, RawIOBase


if hasattr(errno, 'WSAEWOULDBLOCK'):
    _blockers = set((errno.EAGAIN, errno.EWOULDBLOCK, errno.WSAEWOULDBLOCK))
else:
    _blockers = set((errno.EAGAIN, errno.EWOULDBLOCK))


def io_blocked(err):
    """Returns True if IO operation is blocked

    err
        An IOError exception (or similar object with errno attribute).

    Bear in mind that EAGAIN and EWOULDBLOCK are not necessarily the
    same value and that when running under windows WSAEWOULDBLOCK may be
    raised instead.  This function removes this complexity making it
    easier to write cross platform non-blocking IO code."""
    return err.errno in _blockers


if hasattr(errno, 'WSAETIMEDOUT'):
    _timeouts = set((errno.ETIMEDOUT, errno.WSAETIMEDOUT))
else:
    _timeouts = set((errno.ETIMEDOUT, ))


def io_timedout(err):
    """Returns True if an IO operation timed out

    err
        An IOError exception (or similar object with errno attribute).

    Tests for ETIMEDOUT and when running under windows WSAETIMEDOUT
    too."""
    return err.errno in _timeouts


class BufferedStreamWrapper(RawIOBase):

    """A buffered wrapper for file-like objects.

    src
        A file-like object, we only require a read method

    buffsize
        The maximum size of the internal buffer

    On construction the src is read until an end of file condition is
    encountered or until buffsize bytes have been read.  EOF is signaled
    by an empty string returned by src's read method.  Instances then
    behave like readable streams transparently reading from the buffer
    or from the remainder of the src as applicable.

    Instances behave differently depending on whether or not the entire
    src is buffered.  If it is they become seekable and set a value for
    the length attribute.  Otherwise they are not seekable and the
    length attribute is None.

    If src is a non-blocking data source and it becomes blocked,
    indicated by read returning None rather than an empty string, then
    the instance reverts to non-seekable behaviour."""

    def __init__(self, src, buffsize=io.DEFAULT_BUFFER_SIZE):
        self.src = src
        self.buff = io.BytesIO()
        self.bsize = 0
        self.overflow = False
        self.length = None
        while True:
            nbytes = buffsize - self.bsize
            if nbytes <= 0:
                # we've run out of buffer space
                self.overflow = True
                break
            data = src.read(nbytes)
            if data is None:
                # blocked, treat as overflow
                self.overflow = True
                break
            elif data:
                self.buff.write(data)
                self.bsize += len(data)
            else:
                # EOF
                break
        self.pos = 0
        self.buff.seek(0)
        if not self.overflow:
            self.length = self.bsize

    def readable(self):
        return True

    def writable(self):
        return False

    def seekable(self):
        return not self.overflow

    def tell(self):
        if self.overflow:
            raise io.UnsupportedOperation
        else:
            return self.pos

    def seek(self, offset, whence=io.SEEK_SET):
        if self.overflow:
            raise io.UnsupportedOperation
        elif whence == io.SEEK_SET:
            self.pos = offset
        elif whence == io.SEEK_CUR:
            self.pos += offset
        elif whence == io.SEEK_END:
            self.pos = self.length + offset
        else:
            raise ValueError("unrecognized whence value in seek: %s" %
                             repr(whence))
        self.buff.seek(self.pos)

    def readinto(self, b):
        if self.pos < self.bsize:
            # read from the buffer
            data = self.buff.read(len(b))
        elif self.overflow:
            # read from the original source
            data = self.src.read(len(b))
            if data is None:
                # handle blocked read
                return None
        else:
            # end of file
            data = b''
        self.pos += len(data)
        b[0:len(data)] = data
        return len(data)

    def peek(self, nbytes):
        """Read up to nbytes without advancing the position

        If the stream is not seekable and we have read past the end of
        the internal buffer then an empty string will be returned."""
        if self.pos < self.bsize:
            data = self.buff.read(nbytes)
            # reset the position of the buffer
            self.buff.seek(self.pos)
            return data
        else:
            return b''


class Pipe(RawIOBase):

    """Buffered pipe for inter-thread communication

    The purpose of this class is to provide a thread-safe buffer to use
    for communicating between two parts of an application that support
    non-blocking io while reducing to a minimum the amount of
    byte-copying that takes place.

    Essentially, write calls with immutable byte strings are simply
    cached without copying (and always succeed) enabling them to be
    passed directly through to the corresponding read operation in
    streaming situations.  However, to improve flow control a canwrite
    method is provided to help writers moderate the amount of data that
    has to be held in the buffer::

        # data producer thread
        while busy:
            wmax = p.canwrite()
            if wmax:
                data = get_at_most_max_bytes(wmax)
                p.write(data)
            else:
                # do something else while the pipe is blocked
                spin_the_beach_ball()

    bsize
        The buffer size, this is used as a guide only.  When writing
        immutable bytes objects to the pipe the buffer size may be
        exceeded as these can simply be cached and returned directly to
        the reader more efficiently than slicing them up just to adhere
        to the buffer size.  However, if the buffer already contains
        bsize bytes all calls to write will block or return None.
        Defaults to io.DEFAULT_BUFFER_SIZE.

    rblocking
        Controls the blocking behaviour of the read end of this pipe.
        True indicates reads may block waiting for data, False that they
        will not and read may return None.  Defaults to True.

    wblocking
        Controls the blocking behaviour of the write end of the this
        pipe. True indicates writes may block waiting for data, False
        that they will not and write may return None.  Defaults to True.

    timeout
        The number of seconds before a blocked read or write operation
        will timeout.  Defaults to None, which indicates 'wait forever'.
        A value of 0 is not the same as placing both ends of the pipe in
        non-blocking mode (though the effect may be similar).

    name
        An optional name to use for this pipe, the name is used when
        raising errors and when logging"""

    def __init__(self, bsize=io.DEFAULT_BUFFER_SIZE,
                 rblocking=True, wblocking=True, timeout=None,
                 name=None):
        #: the name of the pipe
        self.name = name
        # the maximum buffer size, used for flow control, this
        # is not a hard limit
        self.max = bsize
        # buffered strings of bytes
        self.buffer = []
        # the total size of all strings in the buffer
        self.bsize = 0
        # offset into self.buffer[0]
        self.rpos = 0
        # eof indicator
        self._eof = False
        self.rblocking = rblocking
        # an Event that flags the arrival of a reader
        self.rflag = None
        self.wblocking = wblocking
        # timeout duration
        self.timeout = timeout
        # lock for multi-threading
        self.lock = threading.Condition()
        # state values used for monitoring changes
        self.wstate = 0
        self.rstate = 0
        super(Pipe, self).__init__()

    def __repr__(self):
        if self.name:
            return self.name
        else:
            return super(Pipe, self).__repr__()

    def close(self):
        """closed the Pipe

        This implementation works on a 'reader closes' principle.  The
        writer should simply write the EOF marker to the Pipe (see
        :meth:`write_eof`.

        If the buffer still contains data when it is closed a warning is
        logged."""
        # throw away all data
        logging.debug("Pipe.close %s", repr(self))
        with self.lock:
            if self.buffer:
                logging.warning("Pipe.close for %s discarded non-empty buffer",
                                repr(self))
            self.buffer = []
            self.bsize = 0
            self.rpos = 0
            self._eof = True
            # kill anyone waiting
            self.rstate += 1
            self.wstate += 1
            self.lock.notify_all()
        super(Pipe, self).close()
        # if someone is waiting for a reader - wake them as the reader
        # will never come
        if self.rflag is not None:
            self.rflag.set()

    def readable(self):
        """Pipe's are always readable"""
        return True

    def writable(self):
        """Pipe's are always writeable"""
        return True

    def readblocking(self):
        """Returns True if reads may block"""
        return self.rblocking

    def set_readblocking(self, blocking=True):
        """Sets the readblocking mode of the Pipe.

        blocking
            A boolean, defaults to True indicating that reads may
            block."""
        with self.lock:
            self.rblocking = blocking

    def writeblocking(self):
        """Returns True if writes may block"""
        return self.wblocking

    def set_writeblocking(self, blocking=True):
        """Sets the writeblocking mode of the Pipe.

        blocking
            A boolean, defaults to True indicating that writes may
            block."""
        with self.lock:
            self.wblocking = blocking

    def wait(self, timeout, method):
        if timeout is not None:
            tstart = time.time()
        with self.lock:
            while not method():
                if timeout is None:
                    twait = None
                else:
                    twait = (tstart + timeout) - time.time()
                    if twait < 0:
                        logging.warning("Pipe.wait timedout on %s", repr(self))
                        raise IOError(errno.ETIMEDOUT,
                                      os.strerror(errno.ETIMEDOUT),
                                      "pyslet.http.server.Pipe.wait")
                logging.debug("Pipe.wait waiting for %s", repr(self))
                self.lock.wait(twait)

    def empty(self):
        """Returns True if the buffer is currently empty"""
        with self.lock:
            if self.buffer:
                return False
            else:
                return True

    def buffered(self):
        """Returns the number of buffered bytes in the Pipe"""
        with self.lock:
            return self.bsize - self.rpos

    def canwrite(self):
        """Returns the number of bytes that can be written.

        This value is the number of bytes that can be written in a
        single non-blocking call to write.  0 indicates that the pipe's
        buffer is full.  A call to write may accept more than this but
        *the next* call to write will always accept at least this many.

        This class is fully multithreaded so in situations where there
        are multiple threads writing this call is of limited use.

        If called on a pipe that has had the EOF mark written then
        IOError is raised."""
        with self.lock:
            if self.closed or self._eof:
                raise IOError(
                    errno.EPIPE,
                    "canwrite: can't write past EOF on Pipe object")
            wlen = self.max - self.bsize + self.rpos
            if wlen <= 0:
                wlen = 0
                if self.rflag is not None:
                    self.rflag.clear()
            return wlen

    def set_rflag(self, rflag):
        """Sets an Event triggered when a reader is detected.

        rflag
            An Event instance from the threading module.

        The event will be set each time the Pipe is read.  The flag may
        be cleared at any time by the caller but as a convenience it
        will always be cleared when :py:meth:`canwrite` returns 0.

        The purpose of this flag is to allow a writer to use a custom
        event to monitor whether or not the Pipe is ready to be written.
        If the Pipe is full then the writer will want to wait on this
        flag until a reader appears before attempting to write again.
        Therefore, when canwrite indicates that the buffer is full it
        makes sense that the flag is also cleared.

        If the pipe is closed then the event is set as a warning that
        the pipe will never be read.  (The next call to write will
        then fail.)"""
        with self.lock:
            self.rflag = rflag
            if self.closed:
                self.rflag.set()

    def write_wait(self, timeout=None):
        """Waits for the pipe to become writable or raises IOError

        timeout
            Defaults to None: wait forever.  Otherwise the maximum
            number of seconds to wait for."""
        self.wait(timeout, self.canwrite)

    def flush_wait(self, timeout=None):
        """Waits for the pipe to become empty or raises IOError

        timeout
            Defaults to None: wait forever.  Otherwise the maximum
            number of seconds to wait for."""
        self.wait(timeout, self.empty)

    def canread(self):
        """Returns True if the next call to read will *not* block.

        False indicates that the pipe's buffer is empty and that a call
        to read will block.

        Note that if the buffer is empty but the EOF signal has been
        given with :py:meth:`write_eof` then canread returns True! The
        next call to read will not block but return an empty string
        indicating the EOF."""
        with self.lock:
            if self.closed:
                raise IOError(
                    errno.EPIPE, "can't read from a closed Pipe object")
            if self.buffer or self._eof:
                return True
            else:
                return False

    def read_wait(self, timeout=None):
        """Waits for the pipe to become readable or raises IOError

        timeout
            Defaults to None: wait forever.  Otherwise the maximum
            number of seconds to wait for."""
        self.wait(timeout, self.canread)

    def write(self, b):
        """writes data to the pipe

        The implementation varies depending on the type of b.  If b is
        an immutable bytes object then it is accepted even if this
        overfills the internal buffer (as it is not actually copied).
        If b is a bytearray then data is copied, up to the maximum
        buffer size."""
        if self.timeout is not None:
            tstart = time.time()
        with self.lock:
            if self.closed or self._eof:
                raise IOError(errno.EPIPE,
                              "write: can't write past EOF on Pipe object")
            if isinstance(b, memoryview):
                # catch memory view objects here
                b = b.tobytes()
            wlen = self.max - self.bsize + self.rpos
            while wlen <= 0:
                # block on write or return None
                if self.wblocking:
                    if self.timeout is None:
                        twait = None
                    else:
                        twait = (tstart + self.timeout) - time.time()
                        if twait < 0:
                            logging.warning("Pipe.write timed out for %s",
                                            repr(self))
                            raise IOError(errno.ETIMEDOUT,
                                          os.strerror(errno.ETIMEDOUT),
                                          "pyslet.http.server.Pipe.write")
                    logging.debug("Pipe.write waiting for %s", repr(self))
                    self.lock.wait(twait)
                    # check for eof again!
                    if self.closed or self._eof:
                        raise IOError(errno.EPIPE,
                                      "write: EOF or pipe closed after wait")
                    # recalculate the writable space
                    wlen = self.max - self.bsize + self.rpos
                else:
                    return None
            if isinstance(b, bytes):
                nbytes = len(b)
                if nbytes:
                    self.buffer.append(b)
                    self.bsize += nbytes
                    self.wstate += 1
                    self.lock.notify_all()
                return nbytes
            elif isinstance(b, bytearray):
                nbytes = len(b)
                if nbytes > wlen:
                    nbytes = wlen
                    # partial copy, creates transient bytearray :(
                    self.buffer.append(bytes(b[:nbytes]))
                else:
                    self.buffer.append(bytes(b))
                self.bsize += nbytes
                self.wstate += 1
                self.lock.notify_all()
                return nbytes
            else:
                raise TypeError(repr(type(b)))

    def write_eof(self):
        """Writes the EOF flag to the Pipe

        Any waiting readers are notified and will wake to process the
        Pipe.  After this call the Pipe will not accept any more data."""
        with self.lock:
            self._eof = True
            self.wstate += 1
            self.lock.notify_all()

    def flush(self):
        """flushes the Pipe

        The intention of flush to push any written data out to the
        destination, in this case the thread that is reading the data.

        In write-blocking mode this call will wait until the buffer is
        empty, though if the reader is idle for more than
        :attr:`timeout` seconds then it will raise IOError.

        In non-blocking mode it simple raises IOError with EWOULDBLOCK
        if the buffer is not empty.

        Given that flush is called automatically by :meth:`close` for
        classes that inherit from the base io classes our implementation
        of close discards the buffer rather than risk an exception."""
        if self.timeout is not None:
            tstart = time.time()
        with self.lock:
            blen = self.bsize - self.rpos
            while self.buffer:
                if self.wblocking:
                    if self.timeout is None:
                        twait = None
                    else:
                        new_blen = self.bsize - self.rpos
                        if new_blen < blen:
                            # making progress, restart the clock
                            blen = new_blen
                            tstart = time.time()
                        twait = (tstart + self.timeout) - time.time()
                        if twait < 0:
                            logging.warning("Pipe.flush timed out for %s",
                                            repr(self))
                            logging.debug("Pipe.flush found stuck data: %s",
                                          repr(self.buffer))
                            raise IOError(errno.ETIMEDOUT,
                                          os.strerror(errno.ETIMEDOUT),
                                          "pyslet.http.server.Pipe.flush")
                    logging.debug("Pipe.flush waiting for %s", repr(self))
                    self.lock.wait(twait)
                else:
                    raise io.BlockingIOError(
                        errno.EWOULDBLOCK,
                        "Pipe.flush write blocked on %s" % repr(self))

    def readall(self):
        """Overridden to take care of non-blocking behaviour.

        Warning: readall always blocks until it has read EOF, regardless
        of the rblocking status of the Pipe.

        The problem is that, if the Pipe is set for non-blocking reads
        then we seem to have the choice of returning a partial read (and
        failing to signal that some of the data is still in the pipe) or
        raising an error and losing the partially read data.

        Perhaps ideally we'd return None indicating that we are blocked
        from reading the entire stream but this isn't listed as a
        possible return result for io.RawIOBase.readall and it would be
        tricky to implement anyway as we still need to deal with
        partially read data.

        Ultimately the safe choices are raise an error if called on a
        non-blocking Pipe or simply block.  We do the latter on the
        basis that anyone calling readall clearly intends to wait.

        For a deep discussion of the issues around non-blocking behaviour
        see http://bugs.python.org/issue13322"""
        data = []
        with self.lock:
            save_rblocking = self.rblocking
            try:
                self.rblocking = True
                while True:
                    part = self.read(io.DEFAULT_BUFFER_SIZE)
                    if not part:
                        # end of stream
                        return b''.join(data)
                    else:
                        data.append(part)
            finally:
                self.rlocking = save_rblocking

    def _consolidate_buffer(self):
        with self.lock:
            if self.buffer:
                if self.rpos:
                    self.buffer[0] = self.buffer[0][self.rpos:]
                    self.rpos = 0
                self.buffer = [b''.join(self.buffer)]

    def readmatch(self, match=b'\r\n'):
        """Read until a byte string is matched

        match
            A binary string, defaults to CRLF.

        This operation will block if the string is not matched unless
        the buffer becomes full without a match, in which case IOError
        is raised with code ENOBUFS."""
        with self.lock:
            pos = -1
            while pos < 0:
                if self.buffer:
                    # take care of a special case first
                    pos = self.buffer[0].find(match, self.rpos)
                    if pos < 0:
                        # otherwise consolidate the buffer
                        self._consolidate_buffer()
                        pos = self.buffer[0].find(match)  # rpos is now 0
                if pos >= 0:
                    src = self.buffer[0]
                    result = src[self.rpos:pos + len(match)]
                    self.rpos += len(result)
                    if self.rpos >= len(src):
                        # discard src
                        self.buffer = self.buffer[1:]
                        self.bsize = self.bsize - len(src)
                        self.rpos = 0
                    self.rstate += 1
                    # success, set the reader flag
                    if self.rflag is not None:
                        self.rflag.set()
                    self.lock.notify_all()
                    return result
                else:
                    if self._eof:
                        return b''
                    # not found, should we block?
                    if self.canwrite():
                        # no match, but the buffer is not full so
                        # set the reader flag to indicate that we
                        # are now waiting to accept data.
                        if self.rflag is not None:
                            self.rflag.set()
                        if self.rblocking:
                            # we wait for something to happen on the
                            # Pipe hopefully a write operation!
                            cstate = self.wstate
                            logging.debug("Pipe.readmatch waiting for %s",
                                          repr(self))
                            self.lock.wait(self.timeout)
                            if self.wstate == cstate:
                                logging.warning(
                                    "Pipe.readmatch timed out for %s",
                                    repr(self))
                                raise IOError(
                                    errno.ETIMEDOUT,
                                    os.strerror(errno.ETIMEDOUT),
                                    "pyslet.http.server.Pipe.readmatch")
                            # go round the loop again
                        else:
                            # non-blocking readmatch returns None
                            return None
                    else:
                        # we can't write so no point in waiting
                        raise IOError(errno.ENOBUFS,
                                      os.strerror(errno.ENOBUFS),
                                      "pyslet.http.server.Pipe.readmatch")

    def read(self, nbytes=-1):
        """read data from the pipe

        May return fewer than nbytes if the result can be returned
        without copying data.  Otherwise :meth:`readinto` is used."""
        if nbytes < 0:
            return self.readall()
        else:
            with self.lock:
                if self.buffer and self.rpos == 0:
                    # take care of one special case
                    src = self.buffer[0]
                    if len(src) <= nbytes:
                        self.buffer = self.buffer[1:]
                        self.bsize = self.bsize - len(src)
                        self.rstate += 1
                        # successful read
                        if self.rflag is not None:
                            self.rflag.set()
                        self.lock.notify_all()
                        return src
                b = bytearray(nbytes)
                nbytes = self.readinto(b)
                if nbytes is None:
                    return None
                else:
                    return bytes(b[:nbytes])

    def readinto(self, b):
        """Reads data from the Pipe into a bytearray.

        Returns the number of bytes read.  0 indicates EOF, None
        indicates an operation that would block in a Pipe that is
        non-blocking for read operations.  May return fewer bytes than
        would fit into the bytearray as it returns as soon as it has at
        least some data."""
        if self.timeout is not None:
            tstart = time.time()
        with self.lock:
            nbytes = len(b)
            # we're now reading
            if self.rflag is not None:
                self.rflag.set()
            while not self.buffer:
                if self._eof:
                    return 0
                elif self.rblocking:
                    if self.timeout is None:
                        twait = None
                    else:
                        twait = (tstart + self.timeout) - time.time()
                        if twait < 0:
                            logging.warning("Pipe.read timed out for %s",
                                            repr(self))
                            raise IOError(errno.ETIMEDOUT,
                                          os.strerror(errno.ETIMEDOUT),
                                          "pyslet.http.server.Pipe.read")
                    logging.debug("Pipe.read waiting for %s", repr(self))
                    self.lock.wait(twait)
                else:
                    return None
            src = self.buffer[0]
            rlen = len(src) - self.rpos
            if rlen < nbytes:
                nbytes = rlen
            if nbytes:
                b[:nbytes] = src[self.rpos:self.rpos + nbytes]
                self.rpos += nbytes
            if self.rpos >= len(src):
                # discard src
                self.buffer = self.buffer[1:]
                self.bsize = self.bsize - len(src)
                self.rpos = 0
            if nbytes:
                self.rstate += 1
                self.lock.notify_all()
            return nbytes



#! /usr/bin/env python
"""This module implements the QTI 2.1 specification defined by IMS GLC
"""

import warnings

from .qtiv2.core import *           # noqa
from .qtiv2.variables import *      # noqa
from .qtiv2.expressions import *    # noqa
from .qtiv2.processing import *     # noqa
from .qtiv2.content import *        # noqa
from .qtiv2.interactions import *   # noqa
from .qtiv2.items import *          # noqa
from .qtiv2.tests import *          # noqa
from .qtiv2.metadata import *       # noqa
from .qtiv2.xml import *            # noqa


warnings.warn(
    "pyslet.imsqtiv2p1 is deprecated, use pyslet.qtiv2 package instead",
    DeprecationWarning, stacklevel=3)



#! /usr/bin/env python

import binascii
import errno
import io
import logging
import os
import random
import shutil
import sys
import tempfile
import threading

from .py2 import (
    builtins,
    byte,
    dict_keys,
    force_bytes,
    is_unicode,
    is_text,
    join_bytes,
    py2,
    range3,
    SortableMixin,
    to_text,
    ul)


class VirtualFilePath(SortableMixin):

    """Abstract class representing a virtual file system

    Instances represent paths within a file system.  You can't create an
    instance of VirtualFilePath directly, instead you must create
    instances using a class derived from it.  (Do not call the __init__
    method of VirtualFilePath from your derived classes.)

    All instances are created from one or more strings, either byte
    strings or unicode strings, or existing instances. In the case of
    byte strings the encoding is assumed to be the default encoding of
    the file system.  If multiple arguments are given then they are
    joined to make a single path using :py:meth:`join`.

    Instances can be converted to either binary or character strings,
    use :meth:`to_bytes` for the former.  Note that the builtin str
    function returns a binary string in Python 2, not a character string.

    Instances are immutable, and can be used as keys in dictionaries.
    Instances must be from the same file system to be comparable, the
    unicode representation is used.

    An empty path is False, other paths are True.  You can also compare
    a file path with a string (or unicode string) which is first
    converted to a file path instance."""

    fs_name = None
    """The name of the file system, must be overridden by derived classes.

    The purpose of providing a name for a file system is to enable file
    systems to be mapped onto the authority (host) component of a file
    URL."""

    supports_unicode_filenames = False
    """Indicates whether this file system supports unicode file names
    natively. In general, you don't need to worry about this as all
    methods that accept strings will accept either type of string and
    convert to the native representation.

    When creating derived classes you must also override :attr:`sep`,
    :attr:`curdir`, :attr:`pardir`, :attr:`ext`, :attr:`drive_sep`
    (if applicable) and :attr:`empty` with the correct string types."""

    supports_unc = False
    """Indicates whether this file system supports UNC paths.

    UNC paths are of the general form::

        \\\\ComputerName\\SharedFolder\\Resource

    This format is used in Microsoft Windows.  See :meth:`is_unc` for
    details."""

    supports_drives = False
    """Indicates whether this file system supports 'drives', i.e., is
    Windows-like in having drive letters that may prefix paths."""

    codec = 'utf-8'
    """The codec used by this file system

    This codec is used to convert between byte strings and unicode
    strings.  The default is utf-8."""

    sep = b'/'
    """The path separator used by this file system

    This is either a character or byte string, depending on the setting
    of :attr:`supports_unicode_filenames`."""

    curdir = b'.'
    """The path component that represents the current directory"""

    pardir = b'..'
    """The path component that represents the parent directory"""

    ext = b'.'
    """The extension character"""

    drive_sep = b':'
    """The drive separator"""

    empty = b''
    """An empty path string (for use with join)"""

    @classmethod
    def getcwd(cls):
        """Returns an instance representing the working directory."""
        raise NotImplementedError

    @classmethod
    def getcroot(cls):
        """Returns an instance representing the current root.

        UNIX users will find this odd but in other file systems there
        are multiple roots.  Rather than invent an abstract concept of
        the root of roots we just accept that there can be more than
        one.  (We might struggle to perform actions like :meth:`listdir`
        on the root of roots.)

        The current root is determined by stripping back the current
        working directory until it can no longer be split."""
        wd = cls.getcwd()
        drive, root = wd.splitdrive()
        root = wd
        while True:
            head, tail = root.split()
            if root == head:
                # combine the drive with root
                return drive.join(root)
            elif len(head.path) >= len(root.path):
                raise OSError("Root error %s" % to_text(root.path))
            root = head

    @classmethod
    def mkdtemp(cls, suffix="", prefix=""):
        """Creates a temporary directory in the file system

        Returns an instance representing the path to the new directory.

        Similar to Python's tempfile.mkdtemp, like that function the
        caller is responsible for cleaning up the directory, which can
        be done with :py:meth:`rmtree`."""
        raise NotImplementedError

    @classmethod
    def path_str(cls, arg):
        """Converts a single argument to the correct string type

        File systems can use either binary or character strings and we
        convert between them using :attr:`codec`.  This method takes
        either type of string or an existing instance and returns a path
        string of the correct type."""
        if cls.supports_unicode_filenames:
            # native representation is unicode character
            if is_unicode(arg):
                return arg
            elif isinstance(arg, bytes):
                return arg.decode(cls.codec)
            elif isinstance(arg, cls):
                return arg.path
            else:
                raise TypeError
        else:
            # native representation is binary string
            if is_unicode(arg):
                return arg.encode(cls.codec)
            elif isinstance(arg, bytes):
                return arg
            elif isinstance(arg, cls):
                return arg.path
            else:
                raise TypeError

    def __init__(self, *args):
        if self.__class__ is VirtualFilePath:
            raise NotImplementedError
        elif args:
            #: the path, either character or binary string
            self.path = None
            for arg in args:
                new_path = self.path_str(arg)
                if self.path is None or new_path.startswith(self.sep):
                    self.path = new_path
                elif self.path.endswith(self.sep):
                    self.path = self.path + new_path
                else:
                    self.path = self.path + self.sep + new_path
        else:
            self.path = self.path_str('')

    def __str__(self):
        if py2:
            return self.to_bytes()
        else:
            return self.__unicode__()

    def to_bytes(self):
        """Returns the binary string representation of the path."""
        if is_unicode(self.path):
            return self.path.encode(self.codec)
        else:
            return self.path

    def __unicode__(self):
        if is_unicode(self.path):
            return self.path
        else:
            return self.path.decode(self.codec)

    def sortkey(self):
        """Instances are sortable using character strings."""
        return to_text(self)

    def join(self, *components):
        """Returns a new instance by joining path components

        Starting with the current instance, this method appends each
        component, returning a new instance representing the joined
        path. If components contains an absolute path then previous
        components, including the instance's path, are discarded.

        For details see Python's os.path.join function.

        For the benefit of derived classes a default implementation is
        provided."""
        # split the drive
        drive, tail = self.splitdrive()
        if tail.path:
            new_path = [tail.path]
            add_sep = not tail.path.endswith(self.sep)
        else:
            new_path = []
            add_sep = False
        for component in components:
            new_component = self.__class__(component)
            if self.supports_drives:
                # check for a drive letter
                new_drive, tail = new_component.splitdrive()
                if new_drive:
                    drive = new_drive
                    new_path = []
                    add_sep = False
                    new_component = tail
            if not new_component:
                # an empty component, add a separator (perhaps)
                if add_sep:
                    new_path.append(self.sep)
                    add_sep = False
                continue
            if new_component.isabs():
                new_path = [new_component.path]
                add_sep = not new_component.path.endswith(self.sep)
            else:
                if add_sep:
                    new_path.append(self.sep)
                new_path.append(new_component.path)
                add_sep = not new_component.path.endswith(self.sep)
        if drive:
            new_path[0:0] = [drive.path]
        return self.__class__(self.path_str('').join(new_path))

    def split(self):
        """Splits a path

        Returns a tuple of two instances (head, tail) where tail is the
        last path component and head is everything leading up to it.

        For details see Python's os.path.split."""
        drive, tail = self.splitdrive()
        pos = tail.path.rfind(self.sep)
        if pos >= 0:
            head_path = tail.path[0:pos]
            if head_path:
                # handle special case for UNC
                if not drive and self.supports_unc and head_path == self.sep:
                    # include the following separator
                    head = self.__class__(self.sep + self.sep)
                else:
                    head = self.__class__(drive.path + head_path)
            else:
                # don't return an empty head if sep is in the path
                head = self.__class__(drive.path + self.sep)
            tail = self.__class__(tail.path[pos + len(self.sep):])
            return head, tail
        else:
            # sep is not in the path - it's all tail
            return self.__class__(), self

    def splitext(self):
        """Splits an extension from a path

        Returns a tuple of (root, ext) where root is an instance
        containing just the root file path and ext is a string of
        *characters* representing the orignal path's extension.

        For details see Python's os.path.splitext."""
        if self.is_dirlike():
            return self, self.path_str('')
        else:
            pos = self.path.rfind(self.ext)
            if pos >= 0:
                return self.__class__(self.path[:pos]), \
                    to_text(self.path[pos:])
            else:
                return self, ""

    def splitdrive(self):
        """Splits a drive designation

        Returns a tuple of two instances (drive, tail) where drive is
        either a drive specification or is empty.

        Default implementation uses the :attr:`drive_sep` to determine
        if the first path component is a drive."""
        if self.supports_drives:
            if len(self.path) > 1 and self.path[1:2] == self.drive_sep:
                return self.__class__(self.path[:2]), \
                    self.__class__(self.path[2:])
        return self.__class__(), self

    def splitunc(self):
        """Splits a UNC path

        Returns a tuple of two instances (mount, path) where mount is an
        instance representing the UNC mount point or an instance
        representing the empty path if this isn't a UNC path.

        Default implementation checks for a double separator at the
        start of the path and at least one more separator."""
        if self.supports_unc:
            if self.path.startswith(self.sep + self.sep):
                host_start = 2 * len(self.sep)
                host_end = self.path.find(self.sep, host_start)
                if host_end > host_start:
                    mount_start = host_end + 1
                    mount_end = self.path.find(self.sep, mount_start)
                    if mount_end > mount_start:
                        # //host/mount/path
                        return self.__class__(self.path[:mount_end]), \
                            self.__class__(self.path[mount_end:])
                    else:
                        # //host/mount - return an empty path
                        return self, self.__class__()
                # else just //host, returned as *path*
        return self.__class__(), self

    def abspath(self):
        """Returns an absolute path instance."""
        return self.getcwd().join(self).normpath()

    def realpath(self):
        """Returns a real path, with any symbolic links removed.

        The default implementation normalises the path using
        :meth:`normpath` and :meth:`normcase`."""
        if self.isabs():
            result = self
        else:
            result = self.getcwd().join(self.path)
        return result.normpath().normcase()

    def normpath(self):
        """Returns a normalised path instance."""
        if not self:
            # empty paths become references to the current dir
            # deal with this special case now
            return self.__class__(self.curdir)
        drive, tail = self.splitdrive()
        components = tail.path.split(self.sep)
        i = 0
        while i < len(components):
            component = components[i]
            if not component:
                # empty path component removed except start and
                # special case of all empty components
                if i == 1 and not components[0]:
                    break
                elif i:
                    del components[i]
                else:
                    i = i + 1
            elif component == self.pardir:
                if i:
                    par = components[i - 1]
                    if par == self.pardir or par == self.curdir:
                        # ../.. and ./.. left unchanged
                        i = i + 1
                    elif par:
                        # x/.. becomes '.'
                        components[i - 1:i + 1] = [self.curdir]
                        i = i - 1
                    else:
                        # /.. goes to /
                        del components[i]
                else:
                    # ../ is ignored
                    i = i + 1
            elif component == self.curdir:
                if i > 0:
                    # something to the left
                    del components[i]
                else:
                    # we'll trim later if there is something to the
                    # right as it could be just a load blanks
                    i = i + 1
            else:
                # just a normal component
                i = i + 1
        if len(components) > 1 and components[0] == self.curdir:
            # take care of ./something
            del components[0]
        elif len(components) == 1 and not components[0]:
            # the root is depicted as '/' not an empty path!
            components.append(components[0])
        return self.__class__(drive.path + self.sep.join(components))

    def normcase(self):
        """Returns a case-normalised path instance.

        The default implementation returns the path unchanged."""
        return self

    def is_unc(self):
        """Returns True if this path is a UNC path.

        UNC paths contain a host designation, a path cannot contain a
        drive specification and also be a UNC path.

        Default implementation calls :meth:`splitunc` and returns True
        if the unc component is non-empty."""
        if self.supports_unc:
            unc, rest = self.splitunc()
            return not unc.is_empty()
        else:
            return False

    def is_single_component(self):
        """Returns True if this path is a single, non-root, component.

        E.g., tests that the path does not contain a slash (it may be
        empty)"""
        return self.sep not in self.path

    def __nonzero__(self):
        return not self.is_empty()

    def __bool__(self):
        return not self.is_empty()

    def is_empty(self):
        """Returns True if this path is empty"""
        return not self.path

    def is_dirlike(self):
        """Returns True if this is a directory-like path.

        E.g., test that the path ends in a slash (last component is
        empty)."""
        return self.path.endswith(self.sep) or \
            self.path.endswith(self.sep + self.curdir)

    def is_root(self):
        """Returns True if this is a root path.

        E.g., tests if it consists of just one or more slashes only (not
        counting any drive specification in file systems that support
        them)."""
        if self.isabs():
            head, tail = self.split()
            return head == self
        else:
            return False

    def isabs(self):
        """Returns True if the path is an absolute path."""
        drive, tail = self.splitdrive()
        return tail.path.startswith(self.sep)

    def stat(self):
        """Return information about the path."""
        raise NotImplementedError

    def exists(self):
        """Returns True if this is existing item in the file system."""
        raise NotImplementedError

    def isfile(self):
        """Returns True if this is a regular file in the file system."""
        raise NotImplementedError

    def isdir(self):
        """Returns True if this is a directory in the file system."""
        raise NotImplementedError

    def open(self, mode="r"):
        """Returns an open file-like object from this path."""
        raise NotImplementedError

    def copy(self, dst):
        """Copies a file to dst path like Python's shutil.copy.

        Note that you can't copy between file system implementations."""
        if not self.isfile():
            raise OSError("%s is not a file" % to_text(self))
        if dst.isdir():
            head, tail = self.split()
            dst = dst.join(tail)
        src_stream = dst_stream = None
        try:
            src_stream = self.open('rb')
            dst_stream = dst.open('wb')
            while True:
                data = src_stream.read(io.DEFAULT_BUFFER_SIZE)
                if not data:
                    break
                while data:
                    nbytes = dst_stream.write(data)
                    data = data[nbytes:]
        finally:
            if src_stream is not None:
                src_stream.close()
            if dst_stream is not None:
                dst_stream.close()

    def move(self, dst):
        """Moves a file to dst path like Python's os.rename."""
        raise NotImplementedError

    def remove(self):
        """Removes a file."""
        raise NotImplementedError

    def listdir(self):
        """List directory contents

        Returns a list containing path instances of the entries in the
        directory."""
        raise NotImplementedError

    def chdir(self):
        """Changes the current working directory to this path"""
        raise NotImplementedError

    def mkdir(self):
        """Creates a new directory at this path.

        If an item at this path already exists OSError is raised.  This
        method ignores any trailing separator."""
        raise NotImplementedError

    def makedirs(self):
        """Recursive directory creation function.

        Like mkdir(), but makes all intermediate-level directories
        needed to contain the leaf directory.

        The default implementation repeatedly uses a combination of
        split and mkdir."""
        cpath = self.realpath()
        if cpath.isdir():
            # nothing to do
            return
        elif cpath.exists():
            raise OSError("%s already exists" % to_text(cpath))
        missing = []
        head, tail = cpath.split()
        if not tail:
            # tail is empty, so ignore
            cpath = head
            head, tail = cpath.split()
        while not head.exists():
            missing[0:0] = [head]
            last_head = head
            head, tail = head.split()
            if last_head == head:
                break
        for dpath in missing:
            dpath.mkdir()
            if dpath == cpath:
                cpath = None
        if cpath is not None:
            cpath.mkdir()

    def walk(self):
        """A generator function that walks the file system

        Similar to os.walk. For each directory in the tree rooted at
        this path (including this path itself), it yields a 3-tuple of::

            (dirpath, dirnames, filenames)

        dirpath is an instance, dirnames and filename are lists of path
        instances."""
        raise NotImplementedError

    def rmtree(self, ignore_errors=False):
        """Removes the tree rooted at this directory

        ignore_errors can be used to ignore any errors from the file
        system."""
        raise NotImplementedError

    def __eq__(self, other):
        return self.path == self.path_str(other)

    def __cmp__(self, other):
        if is_text(other):
            other = self.__class__(other)
        elif self.__class__ is not other.__class__:
            raise ValueError(
                "Can't compare file paths from different file systems")
        return cmp(to_text(self), to_text(other))

    def __hash__(self):
        return hash(self.__class__) | hash(to_text(self))


class MemDirectory(object):

    def __init__(self, path=None):
        self.path = path


class MemFileData(io.BytesIO):

    def __init__(self, initial_bytes, memfile):
        self._memfile = memfile
        io.BytesIO.__init__(self, initial_bytes)

    def close(self):
        self.flush()
        self._memfile.data = self.getvalue()
        super(MemFileData, self).close()


class MemFile(object):

    def __init__(self, path=None):
        self.path = path
        self.stream = None
        self.data_stream = None
        self.data = b''

    def is_open(self):
        return self.data_stream is not None and not self.data_stream.closed

    def open(self, mode='r'):
        if self.is_open():
            raise IOError(errno.EPERM, os.strerror(errno.EPERM),
                          "pyslet.vfs.MemFile.open: file already open")
        if mode[0] == 'r':
            readable = True
            writable = '+' in mode
            truncate = False
            append = False
        elif mode[0] == 'w':
            readable = '+' in mode
            writable = True
            truncate = True
            append = False
        elif mode[0] == 'a':
            readable = '+' in mode
            writable = True
            truncate = False
            append = True
        else:
            raise ValueError("Bad mode: %s" % mode)
        if truncate:
            self.data = b''
        # create the base object
        self.data_stream = MemFileData(self.data, self)
        if readable and writable:
            stream = io.BufferedRandom(self.data_stream)
        elif writable:
            stream = io.BufferedWriter(self.data_stream)
        else:
            stream = io.BufferedReader(self.data_stream)
        if append:
            stream.seek(len(self.data))
        if 'b' in mode:
            # binary mode, we're done
            self.stream = stream
        else:
            if 'U' in mode:
                newline = None
            else:
                newline = ''
            self.stream = io.TextIOWrapper(stream, newline=newline)
        return self.stream


class MemFilePath(VirtualFilePath):

    fs_name = "memfs.pyslet.org"
    """Set to memfs.pyslet.org

    This class is registered when the module is loaded enabling
    URLs of the form::

        file://memfs.pyslet.org/dir/file.ext"""

    supports_unicode_filenames = True
    """File names are treated as character (unicode) strings."""

    sep = ul("/")

    curdir = ul(".")

    pardir = ul("..")

    ext = ul(".")

    drive_sep = ul(':')

    empty = ul("")

    _wd = None

    _lock = threading.RLock()

    _fsdir = {}

    @classmethod
    def getcwd(cls):
        """Returns an instance representing the working directory."""
        if cls._wd is None:
            cls._wd = cls('/')
        return cls._wd

    def chdir(self):
        """Sets the current working directory"""
        if self.isabs():
            self.__class__._wd = self
        else:
            self.__class__._wd = self.abspath()

    @classmethod
    def mkdtemp(cls, suffix="", prefix=""):
        """Creates a temporary directory in the file system

        Returns an instance representing the path to the new directory.

        Similar to Python's tempfile.mkdtemp, like that function the
        caller is responsible for cleaning up the directory, which can
        be done with :py:meth:`rmtree`."""
        tmp = cls("/tmp")
        # create a random string of hex digits
        i = 0
        while i < 256:
            i += 1
            try:
                stem = os.urandom(10)
            except NotImplementedError:
                stem = []
                for i in range3(10):
                    stem.append(byte(random.getrandbits(8)))
                stem = join_bytes(stem)
            stem = binascii.hexlify(stem).lower()
            name = cls.path_str(prefix) + cls.path_str(stem) + \
                cls.path_str(suffix)
            if cls.sep in name:
                raise ValueError
            tmp = tmp.join(name)
            # now hold the file system lock
            with cls._lock:
                if not tmp.exists():
                    tmp.makedirs()
                    return tmp
        logging.error("MemFilePath.mkdtemp failed in 256 attempts")
        raise RuntimeError

    def __init__(self, *args):
        super(MemFilePath, self).__init__(*args)

    def exists(self):
        """Returns True if this is an existing item in the file system."""
        # step 1: canonicalise the path
        cpath = self.realpath()
        with self._lock:
            return cpath in self._fsdir

    def isfile(self):
        """Returns True if this is a file in the file system."""
        if self.is_dirlike():
            return False
        cpath = self.realpath()
        with self._lock:
            item = self._fsdir.get(cpath, None)
            return isinstance(item, MemFile)

    def isdir(self):
        """Returns True if this is a directory in the file system."""
        cpath = self.realpath()
        with self._lock:
            item = self._fsdir.get(cpath, None)
            return isinstance(item, MemDirectory)

    def open(self, mode="r"):
        """Returns an open file-like object from this path."""
        if not self or self.is_dirlike():
            # can't open empty paths, or directory-like paths
            raise OSError("%s cannot be opened as a file" % self)
        cpath = self.realpath()
        with self._lock:
            item = self._fsdir.get(cpath, None)
            if item is None:
                item = MemFile(cpath)
                self._fsdir[cpath] = item
            if isinstance(item, MemFile):
                return item.open(mode)
            else:
                # must be a directory, can't open one of those
                raise OSError("%s is a directory" % self)

    def mkdir(self):
        """Creates a new directory at this path."""
        cpath = self.realpath()
        head, tail = cpath.split()
        if not tail:
            # empty last component, repeat
            cpath = head
            head, tail = cpath.split()
        with self._lock:
            if head not in self._fsdir:
                if head.is_root():
                    # auto create the root if missing
                    self._fsdir[head] = MemDirectory(head)
                    if cpath == head:
                        return
                else:
                    raise OSError("%s does not exist" % to_text(head))
            if cpath in self._fsdir:
                raise OSError("%s already exists" % to_text(cpath))
            self._fsdir[cpath] = MemDirectory(cpath)

    def listdir(self):
        """List directory contents"""
        cpath = self.realpath()
        items = []
        with self._lock:
            dir = self._fsdir.get(cpath, None)
            if not isinstance(dir, MemDirectory):
                raise OSError("%s is not a directory" % to_text(cpath))
            if not cpath.is_root():
                prefix = cpath.path + self.sep
            else:
                prefix = cpath.path
            for path in dict_keys(self._fsdir):
                if not path.path.startswith(prefix):
                    continue
                head, tail = path.split()
                if head == cpath:
                    if tail and tail.is_single_component():
                        # catch case of empty tail ('/')
                        items.append(tail)
        return items

    def remove(self):
        """Removes a file."""
        if self.isfile():
            cpath = self.realpath()
            with self._lock:
                item = self._fsdir[cpath]
                if isinstance(item, MemFile):
                    if item.is_open():
                        raise IOError(
                            errno.EPERM, os.strerror(errno.EPERM),
                            "pyslet.vfs.MemFile.remove: file is open")
                    else:
                        del self._fsdir[cpath]
        elif self.isdir():
            raise OSError("%s is a directory" % self)
        else:
            raise OSError("%s no such file" % self)

    def rmtree(self, ignore_errors=False):
        """Removes the tree rooted at this directory"""
        cpath = self.realpath()
        with self._lock:
            dir = self._fsdir.get(cpath, None)
            if not isinstance(dir, MemDirectory):
                if ignore_errors:
                    return
                else:
                    raise OSError("%s is not a directory" % to_text(cpath))
            if not cpath.is_root():
                prefix = cpath.path + self.sep
            else:
                prefix = cpath.path
            dirs = []
            files = []
            open_files = False
            for path in dict_keys(self._fsdir):
                if not path.path.startswith(prefix):
                    continue
                item = self._fsdir[path]
                if isinstance(item, MemDirectory):
                    dirs.append(path)
                else:
                    # it must be a file object
                    if item.is_open():
                        # file is open, can't be removed
                        if not ignore_errors:
                            raise OSError("%s is open" % to_text(path))
                        else:
                            # don't remove any directories
                            open_files = True
                    else:
                        files.append(path)
            for fpath in files:
                del self._fsdir[fpath]
            if not open_files:
                for dpath in dirs:
                    del self._fsdir[dpath]


class OSFilePath(VirtualFilePath):

    """A concrete implementation mapping to Python's os modules

    In most cases the methods map straightforwardly to functions in os
    and os.path."""

    fs_name = ""
    """An empty string.

    The file system name affects the way URIs are interpreted, an empty
    string is consistent with the use of file:/// to reference the
    local file system."""

    supports_unicode_filenames = os.path.supports_unicode_filenames
    """Copied from os.path

    That means you won't know ahead of time whether paths are expected
    as binary or unicode strings.  In most cases it won't matter as the
    methods will convert as appropriate but it does affect the type of
    the static path constants defined below."""

    supports_unc = hasattr(os.path, 'splitunc')
    """Automatically determined from os.path

    Tests if os.path has defined splitunc."""

    supports_drives = not (os.sep in os.path.join("C:", "foo"))
    """Automatically determined

    The method chosen is straight out of the documentation for os.path.
    We join the segments "C:" and "foo" and check to see if the result
    contains the path separator or not."""

    #: as returned by sys.getfilesystemencoding()
    codec = sys.getfilesystemencoding()

    #: copied from os.sep
    sep = os.sep if os.path.supports_unicode_filenames else \
        force_bytes(os.sep)

    #: copied from os.curdir
    curdir = os.curdir if os.path.supports_unicode_filenames else \
        force_bytes(os.curdir)

    #: copied from os.pardir
    pardir = os.pardir if os.path.supports_unicode_filenames else \
        force_bytes(os.pardir)

    #: copied from os.extsep
    ext = os.extsep if os.path.supports_unicode_filenames else \
        force_bytes(os.extsep)

    drive_sep = ul(":") if os.path.supports_unicode_filenames else b":"
    """always set to ':'

    Correctly set to either binary or character string depending on the
    setting of :attr:`supports_unicode_filenames`."""

    empty = ul("") if os.path.supports_unicode_filenames else b""
    """Set to the empty string

    Uses either a binary or character string depending on the setting of
    :attr:`supports_unicode_filenames`."""

    @classmethod
    def getcwd(cls):
        return cls(os.getcwd())

    @classmethod
    def mkdtemp(cls, suffix="", prefix=""):
        return cls(tempfile.mkdtemp(suffix, prefix))

    def __init__(self, *path):
        if path:
            self.path = os.path.join(*map(self.path_str, path))
        else:
            self.path = self.path_str('')

    def join(self, *components):
        return OSFilePath(self, *components)

    def split(self):
        head, tail = os.path.split(self.path)
        return OSFilePath(head), OSFilePath(tail)

    def splitext(self):
        root, ext = os.path.splitext(self.path)
        return OSFilePath(root), to_text(ext)

    def splitdrive(self):
        drive, path = os.path.splitdrive(self.path)
        return OSFilePath(drive), OSFilePath(path)

    def splitunc(self):
        if self.supports_unc:
            unc, path = os.path.splitunc(self.path)
            return OSFilePath(unc), OSFilePath(path)
        else:
            return None, self

    def abspath(self):
        return OSFilePath(os.path.abspath(self.path))

    def realpath(self):
        return OSFilePath(os.path.realpath(self.path))

    def normpath(self):
        return OSFilePath(os.path.normpath(self.path))

    def normcase(self):
        return OSFilePath(os.path.normcase(self.path))

    def is_single_component(self):
        return not (self.sep in self.path)

    def is_empty(self):
        return not len(self.path)

    def isabs(self):
        return os.path.isabs(self.path)

    def stat(self):
        return os.stat(self.path)

    def exists(self):
        return os.path.exists(self.path)

    def isfile(self):
        return os.path.isfile(self.path)

    def isdir(self):
        return os.path.isdir(self.path)

    def open(self, mode="r"):
        return open(self.path, mode)

    def copy(self, dst):
        if not isinstance(dst, self.__class__):
            raise ValueError("Can't copy across file system implementations")
        else:
            shutil.copy(self.path, dst.path)

    def move(self, dst):
        if not isinstance(dst, self.__class__):
            raise ValueError("Can't move across file system implementations")
        else:
            os.rename(self.path, dst.path)

    def remove(self):
        os.remove(self.path)

    def listdir(self):
        return map(lambda x: OSFilePath(x), os.listdir(self.path))

    def chdir(self):
        os.chdir(self.path)

    def mkdir(self):
        os.mkdir(self.path)

    def makedirs(self):
        os.makedirs(self.path)

    def walk(self):
        def wrap(fname):
            return OSFilePath(fname)
        for dirpath, dirnames, filenames in os.walk(self.path):
            yield wrap(dirpath), map(wrap, dirnames), map(wrap, filenames)

    def rmtree(self, ignore_errors=False):
        shutil.rmtree(self.path, ignore_errors)


stat_pass = os.stat
open_pass = builtins.open


def stat_hook(path, *args, **kwargs):
    if isinstance(path, VirtualFilePath):
        return path.stat()
    else:
        return stat_pass(path, *args, **kwargs)


def open_hook(path, *params):
    if isinstance(path, VirtualFilePath):
        return path.open(*params)
    else:
        return open_pass(path, *params)


class ZipHooks(object):

    """Context manager for compatibility with zipfile

    The zipfile module allows you to write either a string or the
    contents of a named file to a zip archive.  This class
    monkey-patches the builtin open function and os.stat with versions
    that support :class:`VirtualFilePath` objects allowing us to copy
    the contents of a virtual represented file path directly to a zip
    archive without having to load it into memory first.

    For more information on this approach see this `blog post`__.

    ..  __:
        http://swl10.blogspot.co.uk/2012/12/writing-stream-to-zipfile-in-python.html

    This implementation uses a lock on the class attributes to ensure
    thread safety.

    As currently implemented, Pyslet does not contain a full
    implementation of :class:`VirtualFilePath` so this class is provided
    in readiness for a more comprehensive implementation based on
    :class:`pyslet.blockstore.StreamStore`."""
    hookCount = 0
    lock = threading.RLock()

    def __init__(self):
        with self.lock:
            if not ZipHooks.hookCount:
                os.stat = stat_hook
                builtins.open = open_hook
            ZipHooks.hookCount += 1
            self.hooked = True

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.unhook()

    def unhook(self):
        with self.lock:
            if self.hooked:
                ZipHooks.hookCount -= 1
                if not ZipHooks.hookCount:
                    os.stat = stat_pass
                    builtins.open = open_pass
                self.hooked = False
            else:
                raise RuntimeError("ZipHooks already unhooked")

fsRegister = {}


def register_file_system(fs):
    fsRegister[fs.fs_name] = fs


def get_file_system_by_name(fs_name):
    return fsRegister.get(fs_name, None)

defaultFS = OSFilePath
register_file_system(OSFilePath)
register_file_system(MemFilePath)



#! /usr/bin/env python

import random
import math

from . import core
from . import variables
from ..pep8 import old_method
from ..py2 import (
    dict_keys,
    dict_values,
    range3,
    uempty)
from ..xml import structures as xml
from ..xml import xsdatatypes as xsi


class Expression(core.QTIElement):

    """Abstract class for all expression elements."""

    @old_method('Evaluate')
    def evaluate(self, state):
        """Evaluates this expression in the context of the session *state*."""
        raise NotImplementedError("Evaluation of %s" % self.__class__.__name__)

    def integer_or_template_ref(self, state, value):
        """Given a value of type integerOrTemplateRef this method returns the
        corresponding integer by looking up the value, if necessary, in
        *state*.  If value is a variable reference to a variable with NULL
        value then None is returned."""
        ref = core.GetTemplateRef(value)
        if ref:
            if state.IsTemplate(ref):
                v = state[ref]
                if isinstance(v, variables.IntegerValue):
                    return v.value
                else:
                    raise core.ProcessingError(
                        "Bad reference: %s is not an integer variable" % ref)
            else:
                raise core.ProcessingError(
                    "Bad reference: %s is not a template variable" % ref)
        else:
            return xsi.integer_from_str(value)

    def float_or_template_ref(self, state, value):
        """Given a value of type floatOrTemplateRef this method returns the
        corresponding float by looking up the value, if necessary, in
        *state*.  If value is a variable reference to a variable with NULL
        value then None is returned."""
        ref = core.GetTemplateRef(value)
        if ref:
            if state.IsTemplate(ref):
                v = state[ref]
                if isinstance(v, variables.FloatValue):
                    return v.value
                else:
                    raise core.ProcessingError(
                        "Bad reference: %s is not a float variable" % ref)
            else:
                raise core.ProcessingError(
                    "Bad reference: %s is not a template variable" % ref)
        else:
            return xsi.double_from_str(value)

    def string_or_template_ref(self, state, value):
        """Given a value of type stringOrTemplateRef this method returns the
        corresponding string by looking up the value, if necessary, in
        *state*.  If value is a variable reference to a variable with NULL
        value then None is returned.  Note that unlike the integer and float
        expansions this expansion will not raise an error if *value* is a
        syntactically valid reference to a non-existent template variable, as
        per this condition in the specification.

                "if a string attribute appears to be a reference to a template
                variable but there is no variable with the given name it should
                be treated simply as string value"
        """
        ref = core.GetTemplateRef(value)
        if ref:
            if state.IsTemplate(ref):
                v = state[ref]
                if isinstance(v, variables.StringValue):
                    return v.value
                else:
                    raise core.ProcessingError(
                        "Bad reference: %s is not a string variable" % ref)
            else:
                return value
        else:
            return value


class BaseValue(Expression):

    """The simplest expression returns a single value from the set defined by
    the given baseType
    ::

            <xsd:attributeGroup name="baseValue.AttrGroup">
                    <xsd:attribute name="baseType" type="baseType.Type"
                    use="required"/>
            </xsd:attributeGroup>

            <xsd:complexType name="baseValue.Type">
                    <xsd:simpleContent>
                            <xsd:extension base="xsd:string">
                                    <xsd:attributeGroup
                                    ref="baseValue.AttrGroup"/>
                            </xsd:extension>
                    </xsd:simpleContent>
            </xsd:complexType>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'baseValue')
    XMLATTR_baseType = (
        'baseType',
        variables.BaseType.from_str_lower,
        variables.BaseType.to_str)
    XMLCONTENT = xml.XMLMixedContent

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.baseType = variables.BaseType.string

    def evaluate(self, state):
        return variables.SingleValue.NewValue(self.baseType, self.get_value())


class Variable(Expression):

    """This expression looks up the value of an itemVariable that has been
    declared in a corresponding variableDeclaration or is one of the built-in
    variables::

            <xsd:attributeGroup name="variable.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                    use="required"/>
                    <xsd:attribute name="weightIdentifier"
                    type="identifier.Type" use="optional"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'variable')
    XMLATTR_identifier = ('identifier', core.ValidateIdentifier, lambda x: x)
    XMLATTR_weightIdentifier = (
        'weightIdentifier', core.ValidateIdentifier, lambda x: x)
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.identifier = ''
        self.weightIdentifier = None

    def evaluate(self, state):
        try:
            return state[self.identifier]
        except KeyError:
            raise core.ProcessingError(
                "%s has not been declared" % self.identifier)


class Default(Expression):

    """This expression looks up the declaration of an itemVariable and returns
    the associated defaultValue or NULL if no default value was declared::

            <xsd:attributeGroup name="default.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                    use="required"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'default')
    XMLATTR_identifier = ('identifier', core.ValidateIdentifier, lambda x: x)
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.identifier = ''

    def evaluate(self, state):
        try:
            state.GetDeclaration(self.identifier)
            return state[self.identifier + ".DEFAULT"]
        except KeyError:
            raise core.ProcessingError(
                "%s has not been declared" % self.identifier)


class Correct(Expression):

    """This expression looks up the declaration of a response variable and
    returns the associated correctResponse or NULL if no correct value was
    declared::

            <xsd:attributeGroup name="correct.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                    use="required"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'correct')
    XMLATTR_identifier = ('identifier', core.ValidateIdentifier, lambda x: x)
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.identifier = ''

    def evaluate(self, state):
        try:
            d = state.GetDeclaration(self.identifier)
            if isinstance(d, variables.ResponseDeclaration):
                return state[self.identifier + ".CORRECT"]
            elif state.IsResponse(self.identifier):
                raise core.ProcessingError(
                    "Can't get the correct value of a built-in response %s" %
                    self.identifier)
            else:
                raise core.ProcessingError(
                    "%s is not a response variable" % self.identifier)
        except KeyError:
            raise core.ProcessingError(
                "%s has not been declared" % self.identifier)


class MapResponse(Expression):

    """This expression looks up the value of a response variable and then
    transforms it using the associated mapping, which must have been declared.
    The result is a single float::

            <xsd:attributeGroup name="mapResponse.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                    use="required"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'mapResponse')
    XMLATTR_identifier = ('identifier', core.ValidateIdentifier, lambda x: x)
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.identifier = ''

    def evaluate(self, state):
        try:
            d = state.GetDeclaration(self.identifier)
            if isinstance(d, variables.ResponseDeclaration):
                if d.Mapping is None:
                    raise core.ProcessingError(
                        "%s has no mapping" % self.identifier)
                return d.Mapping.MapValue(state[self.identifier])
            elif state.IsResponse(self.identifier):
                raise core.ProcessingError(
                    "Can't map built-in response %s" % self.identifier)
            else:
                raise core.ProcessingError(
                    "%s is not a response variable" % self.identifier)
        except KeyError:
            raise core.ProcessingError(
                "%s has not been declared" % self.identifier)


class MapResponsePoint(Expression):

    """This expression looks up the value of a response variable that must be of
    base-type point, and transforms it using the associated areaMapping::

            <xsd:attributeGroup name="mapResponsePoint.AttrGroup">
                    <xsd:attribute name="identifier" type="identifier.Type"
                    use="required"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'mapResponsePoint')
    XMLATTR_identifier = ('identifier', core.ValidateIdentifier, lambda x: x)
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.identifier = ''

    def evaluate(self, state):
        try:
            d = state.GetDeclaration(self.identifier)
            if isinstance(d, variables.ResponseDeclaration):
                if d.baseType is not variables.BaseType.point:
                    raise core.ProcessingError(
                        "%s does not have point type" % self.identifier)
                elif d.AreaMapping is None:
                    raise core.ProcessingError(
                        "%s has no areaMapping" % self.identifier)
                width, height = d.get_stage_dimensions()
                return d.AreaMapping.MapValue(
                    state[
                        self.identifier],
                    width,
                    height)
            elif state.IsResponse(self.identifier):
                raise core.ProcessingError(
                    "Can't map built-in response %s" % self.identifier)
            else:
                raise core.ProcessingError(
                    "%s is not a response variable" % self.identifier)
        except KeyError:
            raise core.ProcessingError(
                "%s has not been declared" % self.identifier)


class Null(Expression):

    """null is a simple expression that returns the NULL value - the null value is
    treated as if it is of any desired baseType
    ::

            <xsd:complexType name="null.Type"/>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'null')
    XMLCONTENT = xml.XMLEmpty

    def evaluate(self, state):
        return variables.Value()


class RandomInteger(Expression):

    """Selects a random integer from the specified range [min,max] satisfying
    min + step * n for some integer n::

            <xsd:attributeGroup name="randomInteger.AttrGroup">
                    <xsd:attribute name="min" type="integerOrTemplateRef.Type"
                    use="required"/>
                    <xsd:attribute name="max" type="integerOrTemplateRef.Type"
                    use="required"/>
                    <xsd:attribute name="step" type="integerOrTemplateRef.Type"
                    use="optional"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'randomInteger')
    XMLATTR_min = 'min'
    XMLATTR_max = 'max'
    XMLATTR_step = 'step'
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.min = "0"
        self.max = None
        self.step = "1"

    def evaluate(self, state):
        min = self.integer_or_template_ref(state, self.min)
        max = self.integer_or_template_ref(state, self.max)
        step = self.integer_or_template_ref(state, self.step)
        return variables.IntegerValue(
            min +
            step *
            random.randint(
                0,
                (max -
                 min) //
                step))


class RandomFloat(Expression):

    """Selects a random float from the specified range [min,max]
    ::

            <xsd:attributeGroup name="randomFloat.AttrGroup">
                    <xsd:attribute name="min" type="floatOrTemplateRef.Type"
                    use="required"/>
                    <xsd:attribute name="max" type="floatOrTemplateRef.Type"
                    use="required"/>
            </xsd:attributeGroup>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'randomFloat')
    XMLATTR_min = 'min'
    XMLATTR_max = 'max'
    XMLCONTENT = xml.XMLEmpty

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.min = "0"
        self.max = None

    def evaluate(self, state):
        min = self.float_or_template_ref(state, self.min)
        max = self.float_or_template_ref(state, self.max)
        # strictly speaking, we can never return max, but due to possible
        # rounding this point is academic
        return variables.FloatValue(min + random.random() * (max - min))


class NOperator(Expression):

    """An abstract class to help implement operators which take multiple
    sub-expressions."""
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.Expression = []

    def get_children(self):
        return iter(self.Expression)

    @old_method('EvaluateChildren')
    def evaluate_children(self, state):
        """Evaluates all child expressions, returning an iterable of
        :py:class:`Value` instances."""
        for e in self.Expression:
            yield e.evaluate(state)


class UnaryOperator(Expression):

    """An abstract class to help implement unary operators."""
    XMLCONTENT = xml.ElementContent

    def __init__(self, parent):
        Expression.__init__(self, parent)
        self.Expression = None

    def get_children(self):
        if self.Expression:
            yield self.Expression


class Multiple(NOperator):

    """The multiple operator takes 0 or more sub-expressions all of which must
    have either single or multiple cardinality::

            <xsd:group name="multiple.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'multiple')

    def evaluate(self, state):
        # We are going to recurse through the sub-expressions evaluating as we
        # go...
        base_type = None
        values = list(self.evaluate_children(state))
        vinput = []
        for v in values:
            if v.baseType is None:
                continue
            if base_type is None:
                base_type = v.baseType
            elif base_type != v.baseType:
                raise core.ProcessingError(
                    "Mixed containers are not allowed: expected %s, found %s" %
                    (variables.BaseType.to_str(base_type),
                     variables.BaseType.to_str(
                        v.baseType)))
            if not v:
                # ignore NULL
                continue
            if v.Cardinality() == variables.Cardinality.single:
                vinput.append(v.value)
            elif v.Cardinality() == variables.Cardinality.multiple:
                # apologies for the obscure code, but this turns {'x':2,'y':3}
                # into ['y', 'y', 'y', 'x', 'x']
                vinput = vinput + list(v.get_values())
            else:
                raise core.ProcessingError(
                    "Ordered or Record values not allowed in Mutiple")
        # finally we have a matching list of input values
        result = variables.MultipleContainer(base_type)
        result.set_value(vinput)
        return result


class Ordered(NOperator):

    """The multiple operator takes 0 or more sub-expressions all of which must
    have either single or multiple cardinality::

            <xsd:group name="ordered.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'ordered')

    def evaluate(self, state):
        # We are going to recurse through the sub-expressions evaluating as we
        # go...
        base_type = None
        values = list(self.evaluate_children(state))
        vinput = []
        for v in values:
            if v.baseType is None:
                continue
            if base_type is None:
                base_type = v.baseType
            elif base_type != v.baseType:
                raise core.ProcessingError(
                    "Mixed containers are not allowed: expected %s, found %s" %
                    (variables.BaseType.to_str(base_type),
                     variables.BaseType.to_str(
                        v.baseType)))
            if not v:
                # ignore NULL
                continue
            if v.Cardinality() == variables.Cardinality.single:
                vinput.append(v.value)
            elif v.Cardinality() == variables.Cardinality.ordered:
                vinput = vinput + list(v.value)
            else:
                raise core.ProcessingError(
                    "Multiple or Record values not allowed in Ordered")
        # finally we have a matching list of input values
        result = variables.OrderedContainer(base_type)
        result.set_value(vinput)
        return result


class ContainerSize(UnaryOperator):

    """The containerSize operator takes a sub-expression with any base-type and
    either multiple or ordered cardinality::

            <xsd:group name="containerSize.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'containerSize')

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        if value.Cardinality() == variables.Cardinality.ordered:
            if value:
                return variables.IntegerValue(len(value.value))
            else:
                return variables.IntegerValue(0)
        elif value.Cardinality() == variables.Cardinality.multiple:
            if value:
                # multiple containers are kept as a mapping to value
                # frequencies
                sum = 0
                for v in dict_values(value.value):
                    sum += v
                return variables.IntegerValue(sum)
            else:
                return variables.IntegerValue(0)
        else:
            raise core.ProcessingError(
                "Ordered or Multiple value required for containerSize")


class IsNull(UnaryOperator):

    """The isNull operator takes a sub-expression with any base-type and
    cardinality
    ::

            <xsd:group name="isNull.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'isNull')

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        if value:
            return variables.BooleanValue(False)
        else:
            return variables.BooleanValue(True)


class Index(UnaryOperator):

    """The index operator takes a sub-expression with an ordered container value
    and any base-type
    ::

            <xsd:attributeGroup name="index.AttrGroup">
                    <xsd:attribute name="n" type="integer.Type"
                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="index.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'index')
    XMLATTR_n = ('n', xsi.integer_from_str, xsi.integer_to_str)

    def __init__(self, parent):
        UnaryOperator.__init__(self, parent)
        self.n = None

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        if value.Cardinality() == variables.Cardinality.ordered:
            result = variables.SingleValue.NewValue(value.baseType)
            if value:
                if self.n < 1:
                    raise core.ProcessingError(
                        "Index requires n>0, found %i" % self.n)
                elif self.n <= len(value.value):
                    result.set_value(value.value[self.n - 1])
            return result
        else:
            # wrong cardinality
            raise core.ProcessingError(
                "Index requires ordered value, found %s" %
                variables.Cardinality.to_str(
                    value.Cardinality()))


class FieldValue(UnaryOperator):

    """The field-value operator takes a sub-expression with a record container
    value. The result is the value of the field with the specified
    fieldIdentifier::

            <xsd:attributeGroup name="fieldValue.AttrGroup">
                    <xsd:attribute name="fieldIdentifier"
                    type="identifier.Type" use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="fieldValue.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'fieldValue')
    XMLATTR_fieldIdentifier = (
        'fieldIdentifier', core.ValidateIdentifier, lambda x: x)

    def __init__(self, parent):
        UnaryOperator.__init__(self, parent)
        self.fieldIdentifier = ''

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        if value.Cardinality() == variables.Cardinality.record:
            if self.fieldIdentifier in value:
                return value[self.fieldIdentifier]
            else:
                return variables.SingleValue()
        else:
            # wrong cardinality
            raise core.ProcessingError(
                "fieldValue requires record value, found %s" %
                variables.Cardinality.to_str(
                    value.Cardinality()))


class Random(UnaryOperator):

    """The random operator takes a sub-expression with a multiple or ordered
    container value and any base-type. The result is a single value randomly
    selected from the container::

            <xsd:group name="random.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'random')

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        if value.Cardinality() == variables.Cardinality.ordered:
            result = variables.SingleValue.NewValue(value.baseType)
            if value:
                # randomly pick one of these values
                result.set_value(random.choice(value.value))
            return result
        elif value.Cardinality() == variables.Cardinality.multiple:
            result = variables.SingleValue.NewValue(value.baseType)
            if value:
                result.set_value(random.choice(list(value.get_values())))
            return result
        elif value.Cardinality() is None:
            return variables.SingleValue()
        else:
            # wrong cardinality
            raise core.ProcessingError(
                "Random requires multiple or ordered value, found %s" %
                variables.Cardinality.to_str(
                    value.Cardinality()))


class Member(NOperator):

    """The member operator takes two sub-expressions which must both have the
    same base-type. The first sub-expression must have single cardinality and
    the second must be a multiple or ordered container. The result is a single
    boolean with a value of true if the value given by the first sub-expression
    is in the container defined by the second sub-expression::

            <xsd:group name="member.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'member')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "Member requires two sub-expressions, found %i" % len(values))
        singleValue, containerValue = values
        if singleValue.baseType is None or containerValue.baseType is None:
            return variables.BooleanValue()
        if singleValue.baseType != containerValue.baseType:
            raise core.ProcessingError(
                "Mismatched base types for member operator")
        if singleValue.baseType == variables.BaseType.duration:
            raise core.ProcessingError(
                "Member operator must not be used on duration values")
        if singleValue and containerValue:
            if singleValue.Cardinality() != variables.Cardinality.single:
                raise core.ProcessingError(
                    "Expected single value, found %s" %
                    variables.Cardinality.to_str(
                        singleValue.Cardinality()))
            if containerValue.Cardinality() == variables.Cardinality.ordered:
                return variables.BooleanValue(
                    singleValue.value in containerValue.value)
            elif (containerValue.Cardinality() ==
                  variables.Cardinality.multiple):
                return variables.BooleanValue(
                    singleValue.value in containerValue.value)
            else:
                raise core.ProcessingError(
                    "Expected ordered or multiple value, found %s" %
                    variables.Cardinality.to_str(
                        containerValue.Cardinality()))
        else:
            return variables.BooleanValue()


class Delete(NOperator):

    """The delete operator takes two sub-expressions which must both have the
    same base-type. The first sub-expression must have single cardinality and
    the second must be a multiple or ordered container. The result is a new
    container derived from the second sub-expression with all instances of the
    first sub-expression removed::

            <xsd:group name="delete.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'delete')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "Delete requires two sub-expressions, found %i" % len(values))
        singleValue, containerValue = values
        if containerValue.baseType is None:
            return variables.Container.NewValue(
                containerValue.Cardinality(),
                singleValue.baseType)
        elif singleValue.baseType is None:
            return variables.Container.NewValue(
                containerValue.Cardinality(),
                containerValue.baseType)
        elif singleValue.baseType != containerValue.baseType:
            raise core.ProcessingError(
                "Mismatched base types for delete operator")
        if singleValue.baseType == variables.BaseType.duration:
            raise core.ProcessingError(
                "Delete operator must not be used on duration values")
        if singleValue.Cardinality() not in (variables.Cardinality.single,
                                             None):
            raise core.ProcessingError(
                "Expected single value, found %s" %
                variables.Cardinality.to_str(
                    singleValue.Cardinality()))
        if containerValue.Cardinality() not in (variables.Cardinality.ordered,
                                                variables.Cardinality.multiple,
                                                None):
            raise core.ProcessingError(
                "Expected ordered or multiple value, found %s" %
                variables.Cardinality.to_str(
                    containerValue.Cardinality()))
        result = variables.Container.NewValue(
            containerValue.Cardinality(), containerValue.baseType)
        if singleValue and containerValue:
            # we treat multiple and ordered the same, less efficient as we
            # enumerate all members
            vresult = []
            for v in containerValue.get_values():
                if singleValue.value != v:
                    vresult.append(v)
            result.set_value(vresult)
            return result
        else:
            return result


class Contains(NOperator):

    """The contains operator takes two sub-expressions which must both have the
    same base-type and cardinality -- either multiple or ordered. The result is
    a single boolean with a value of true if the container given by the first
    sub-expression contains the value given by the second sub-expression and
    false if it doesn't::

            <xsd:group name="contains.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'contains')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "Contains requires two sub-expressions, found %i" %
                len(values))
        v1, v2 = values
        base_type = variables.check_base_types(v1.baseType, v2.baseType)
        cardinality = variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality())
        if base_type == variables.BaseType.duration:
            raise core.ProcessingError(
                "Contains operator must not be used on duration values")
        if cardinality is None or base_type is None:
            return variables.BooleanValue()
        if v1 and v2:
            if cardinality == variables.Cardinality.ordered:
                # we use a naive implementation here as we don't expect large
                # containers
                imax = len(v2.value) - len(v1.value)
                for i in range3(imax + 1):
                    # try a match starting at i
                    match = True
                    for iv1, iv2 in zip(v1.value,
                                        v2.value[i:i + len(v1.value)]):
                        if iv1 != iv2:
                            match = False
                            break
                    if match:
                        return variables.BooleanValue(True)
                return variables.BooleanValue(False)
            elif cardinality == variables.Cardinality.multiple:
                # this is much easier to do
                match = True
                for iv1 in v1.value:
                    if v2.value.get(iv1, 0) < v1.value[iv1]:
                        # no match
                        match = False
                        break
                return variables.BooleanValue(match)
            else:
                raise core.ProcessingError(
                    "Expected ordered or multiple value, found %s" %
                    variables.Cardinality.to_str(cardinality))
        else:
            return variables.BooleanValue()


class SubString(NOperator):

    """The substring operator takes two sub-expressions which must both have an
    effective base-type of string and single cardinality. The result is a
    single boolean with a value of true if the first expression is a substring
    of the second expression and false if it isn't::

            <xsd:attributeGroup name="substring.AttrGroup">
                    <xsd:attribute name="caseSensitive" type="boolean.Type"
                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="substring.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'substring')
    XMLATTR_caseSensitive = (
        'caseSensitive', xsi.boolean_from_str, xsi.boolean_to_str)

    def __init__(self, parent):
        NOperator.__init__(self, parent)
        self.caseSensitive = True

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "substring requires two sub-expressions, found %i" %
                len(values))
        v1, v2 = values
        variables.check_base_types(
            v1.baseType, v2.baseType, variables.BaseType.string)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            if self.caseSensitive:
                return variables.BooleanValue(v1.value in v2.value)
            else:
                return variables.BooleanValue(
                    v1.value.lower() in v2.value.lower())
        else:
            return variables.BooleanValue()


class Not(UnaryOperator):

    """The not operator takes a single sub-expression with a base-type of
    boolean and single cardinality. The result is a single boolean with a value
    obtained by the logical negation of the sub-expression's value::

            <xsd:group name="not.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'not')

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.boolean)
        variables.check_cardinalities(
            value.Cardinality(), variables.Cardinality.single)
        if value:
            return variables.BooleanValue(not value.value)
        else:
            return variables.BooleanValue()


class And(NOperator):

    """The and operator takes one or more sub-expressions each with a base-type
    of boolean and single cardinality. The result is a single boolean which is
    true if all sub-expressions are true and false if any of them are false::

            <xsd:group name="and.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'and')

    def evaluate(self, state):
        # We are going to recurse through the sub-expressions evaluating as we
        # go...
        values = list(self.evaluate_children(state))
        variables.check_cardinalities(variables.Cardinality.single,
                                      *[x.Cardinality() for x in values])
        variables.check_base_types(variables.BaseType.boolean,
                                   *[x.baseType for x in values])
        result = True
        for v in values:
            if v.value is False:
                return variables.BooleanValue(False)
            elif v.value is None:
                result = None
        return variables.BooleanValue(result)


class Or(NOperator):

    """The or operator takes one or more sub-expressions each with a base-type
    of boolean and single cardinality. The result is a single boolean which is
    true if any of the sub-expressions are true and false if all of them are
    false::

            <xsd:group name="or.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'or')

    def evaluate(self, state):
        # We are going to recurse through the sub-expressions evaluating as we
        # go...
        values = list(self.evaluate_children(state))
        variables.check_cardinalities(variables.Cardinality.single,
                                      *[x.Cardinality() for x in values])
        variables.check_base_types(variables.BaseType.boolean,
                                   *[x.baseType for x in values])
        result = False
        for v in values:
            if v.value:
                return variables.BooleanValue(True)
            elif v.value is None:
                result = None
        return variables.BooleanValue(result)


class AnyN(NOperator):

    """The anyN operator takes one or more sub-expressions each with a base-type
    of boolean and single cardinality. The result is a single boolean which is
    true if at least min of the sub-expressions are true and at most max of the
    sub-expressions are true::

            <xsd:attributeGroup name="anyN.AttrGroup">
                    <xsd:attribute name="min" type="integerOrTemplateRef.Type"
                    use="required"/>
                    <xsd:attribute name="max" type="integerOrTemplateRef.Type"
                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="anyN.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'anyN')
    XMLATTR_min = 'min'
    XMLATTR_max = 'max'

    def __init__(self, parent):
        NOperator.__init__(self, parent)
        self.min = None
        self.max = None

    def evaluate(self, state):
        # We are going to recurse through the sub-expressions evaluating as we
        # go...
        min = self.integer_or_template_ref(state, self.min)
        max = self.integer_or_template_ref(state, self.max)
        values = list(self.evaluate_children(state))
        variables.check_cardinalities(variables.Cardinality.single,
                                      *[x.Cardinality() for x in values])
        variables.check_base_types(variables.BaseType.boolean,
                                   *[x.baseType for x in values])
        ntrue = nnull = 0
        for v in values:
            if v.value:
                ntrue += 1
            elif v.value is None:
                nnull += 1
        # The hard part is figuring out what to return!
        if ntrue >= min:
            # could be True
            if ntrue > max:
                # we overshot
                result = False
            elif ntrue + nnull > max:
                # we could still overshoot
                result = None
            else:
                result = True
        elif ntrue + nnull >= min:
            # we could still be True
            result = None
        else:
            result = False
        return variables.BooleanValue(result)


class Match(NOperator):

    """The match operator takes two sub-expressions which must both have the
    same base-type and cardinality. The result is a single boolean with a value
    of true if the two expressions represent the same value and false if they
    do not::

            <xsd:group name="match.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'match')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "Match requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        try:
            return variables.BooleanValue(v1 == v2)
        except variables.NullResult as null:
            return null.value


class StringMatch(NOperator):

    """The stringMatch operator takes two sub-expressions which must have single
    and a base-type of string. The result is a single boolean with a value of
    true if the two strings match::

            <xsd:attributeGroup name="stringMatch.AttrGroup">
                    <xsd:attribute name="caseSensitive" type="boolean.Type"
                    use="required"/>
                    <xsd:attribute name="substring" type="boolean.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="stringMatch.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'stringMatch')
    XMLATTR_caseSensitive = (
        'caseSensitive', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_substring = ('substring', xsi.boolean_from_str, xsi.boolean_to_str)

    def __init__(self, parent):
        NOperator.__init__(self, parent)
        self.caseSensitive = True
        self.substring = None

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "stringMatch requires two sub-expressions, found %i" %
                len(values))
        v1, v2 = values
        variables.check_base_types(
            v1.baseType, v2.baseType, variables.BaseType.string)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            if not self.caseSensitive:
                v1v = v1.value.lower()
                v2v = v2.value.lower()
            else:
                v1v = v1.value
                v2v = v2.value
            if self.substring:
                return variables.BooleanValue(v1v in v2v)
            else:
                return variables.BooleanValue(v1v == v2v)
        else:
            return variables.BooleanValue()


class PatternMatch(UnaryOperator):

    """The patternMatch operator takes a sub-expression which must have single
    cardinality and a base-type of string. The result is a single boolean with
    a value of true if the sub-expression matches the regular expression given
    by pattern and false if it doesn't::

            <xsd:attributeGroup name="patternMatch.AttrGroup">
                    <xsd:attribute name="pattern"
                    type="stringOrTemplateRef.Type"
                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="patternMatch.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'patternMatch')
    XMLATTR_pattern = 'pattern'

    def __init__(self, parent):
        UnaryOperator.__init__(self, parent)
        self.pattern = uempty

    def evaluate(self, state):
        pattern = self.string_or_template_ref(state, self.pattern)
        value = self.Expression.evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.string)
        variables.check_cardinalities(
            value.Cardinality(), variables.Cardinality.single)
        if value:
            try:
                re = xsi.RegularExpression(pattern)
                return variables.BooleanValue(re.match(value.value))
            except xsi.RegularExpressionError:
                # illegal regular expression results in NULL
                return variables.BooleanValue()
        else:
            return variables.BooleanValue()


class ToleranceMode(xsi.Enumeration):

    """When comparing two floating point numbers for equality it is often
    desirable to have a tolerance to ensure that spurious errors in scoring are
    not introduced by rounding errors. The tolerance mode determines whether
    the comparison is done exactly, using an absolute range or a relative
    range::

            <xsd:simpleType name="toleranceMode.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="absolute"/>
                            <xsd:enumeration value="exact"/>
                            <xsd:enumeration value="relative"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Defines constants for the above modes.  Usage example::

            ToleranceMode.exact

    The default value is exact::

            ToleranceMode.DEFAULT == ToleranceMode.exact

    For more methods see :py:class:`~pyslet.xml.xsdatatypes.Enumeration`"""
    decode = {
        'absolute': 1,
        'exact': 2,
        'relative': 3
    }
    aliases = {
        None: 'exact'
    }


class Equal(NOperator):

    """The equal operator takes two sub-expressions which must both have single
    cardinality and have a numerical base-type. The result is a single boolean
    with a value of true if the two expressions are numerically equal and false
    if they are not::

            <xsd:attributeGroup name="equal.AttrGroup">
                    <xsd:attribute name="toleranceMode"
                                   type="toleranceMode.Type" use="required"/>
                    <xsd:attribute name="tolerance" use="optional">
                            <xsd:simpleType>
                                    <xsd:list
                                    itemType="floatOrTemplateRef.Type"/>
                            </xsd:simpleType>
                    </xsd:attribute>
                    <xsd:attribute name="includeLowerBound" type="boolean.Type"
                    use="optional"/>
                    <xsd:attribute name="includeUpperBound" type="boolean.Type"
                    use="optional"/>
            </xsd:attributeGroup>

            <xsd:group name="equal.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'equal')
    XMLATTR_toleranceMode = (
        'toleranceMode',
        ToleranceMode.from_str_lower,
        ToleranceMode.to_str)
    XMLATTR_tolerance = 'tolerance'
    XMLATTR_includeLowerBound = (
        'includeLowerBound', xsi.boolean_from_str, xsi.boolean_to_str)
    XMLATTR_includeUpperBound = (
        'includeUpperBound', xsi.boolean_from_str, xsi.boolean_to_str)

    def __init__(self, parent):
        NOperator.__init__(self, parent)
        self.toleranceMode = ToleranceMode.exact
        self.tolerance = []
        self.includeLowerBound = True
        self.includeUpperBound = True

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "equal requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            # grab two floats
            v1v = float(v1.value)
            v2v = float(v2.value)
            if self.toleranceMode == ToleranceMode.exact:
                return variables.BooleanValue(v1v == v2v)
            else:
                if len(self.tolerance) == 0:
                    raise core.ProcessingError(
                        "Inexact equal test requires specificed tolerance.")
                t0 = self.float_or_template_ref(state, self.tolerance[0])
                if len(self.tolerance) > 1:
                    t1 = self.float_or_template_ref(state, self.tolerance[1])
                else:
                    t1 = t0
                if self.toleranceMode == ToleranceMode.absolute:
                    lower_bound = v1v - t0
                    upper_bound = v1v + t1
                else:
                    # relative
                    lower_bound = v1v * (1 - t0 / 100.0)
                    upper_bound = v1v * (1 + t1 / 100.0)
                if self.includeUpperBound:
                    if self.includeLowerBound:
                        result = (v2v >= lower_bound and v2v <= upper_bound)
                    else:
                        result = (v2v > lower_bound and v2v <= upper_bound)
                else:
                    if self.includeLowerBound:
                        result = (v2v >= lower_bound and v2v < upper_bound)
                    else:
                        result = (v2v > lower_bound and v2v < upper_bound)
                return variables.BooleanValue(result)
        else:
            return variables.BooleanValue()


class RoundingMode(xsi.Enumeration):

    """Numbers are rounded to a given number of significantFigures or
    decimalPlaces::

            <xsd:simpleType name="roundingMode.Type">
                    <xsd:restriction base="xsd:NMTOKEN">
                            <xsd:enumeration value="decimalPlaces"/>
                            <xsd:enumeration value="significantFigures"/>
                    </xsd:restriction>
            </xsd:simpleType>

    Defines constants for the above modes.  Usage example::

            RoundingMode.decimalPlaces

    The default value is significantFigures::

            RoundingMode.DEFAULT == RoundingMode.significantFigures

    For more methods see :py:class:`~pyslet.xml.xsdatatypes.Enumeration`"""
    decode = {
        'decimalPlaces': 1,
        'significantFigures': 2
    }
    aliases = {
        None: 'significantFigures'
    }


class EqualRounded(NOperator):

    """The equalRounded operator takes two sub-expressions which must both have
    single cardinality and have a numerical base-type. The result is a single
    boolean with a value of true if the two expressions are numerically equal
    after rounding and false if they are not::

            <xsd:attributeGroup name="equalRounded.AttrGroup">
                    <xsd:attribute name="roundingMode" type="roundingMode.Type"
                    use="required"/>
                    <xsd:attribute name="figures"
                                   type="integerOrTemplateRef.Type"
                                   use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="equalRounded.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                                       minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'equalRounded')
    XMLATTR_roundingMode = (
        'roundingMode', RoundingMode.from_str, RoundingMode.to_str)
    XMLATTR_figures = 'figures'

    def __init__(self, parent):
        NOperator.__init__(self, parent)
        self.roundingMode = RoundingMode.DEFAULT
        self.figures = None

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "equalRounded requires two sub-expressions, found %i" %
                len(values))
        v1, v2 = values
        variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            # grab two floats
            v1v = float(v1.value)
            v2v = float(v2.value)
            figures = self.integer_or_template_ref(state, self.figures)
            if self.roundingMode == RoundingMode.decimalPlaces:
                return variables.BooleanValue(
                    xsi.decimal_to_str(
                        v1v,
                        figures,
                        False) == xsi.decimal_to_str(
                        v2v,
                        figures,
                        False))
            else:
                # for sig fig, we need to use the double form
                return variables.BooleanValue(
                    xsi.double_to_str(
                        v1v,
                        figures -
                        1) == xsi.double_to_str(
                        v2v,
                        figures -
                        1))
        else:
            return variables.BooleanValue()


class Inside(UnaryOperator, core.ShapeElementMixin):

    """The inside operator takes a single sub-expression which must have a
    baseType of point. The result is a single boolean with a value of true if
    the given point is inside the area defined by shape and coords. If the
    sub-expression is a container the result is true if any of the points are
    inside the area::

            <xsd:attributeGroup name="inside.AttrGroup">
                    <xsd:attribute name="shape" type="shape.Type"
                    use="required"/>
                    <xsd:attribute name="coords" type="coords.Type"
                    use="required"/>
            </xsd:attributeGroup>

            <xsd:group name="inside.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'inside')

    def __init__(self, parent):
        UnaryOperator.__init__(self, parent)
        core.ShapeElementMixin.__init__(self)

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.point)
        if value.Cardinality() == variables.Cardinality.single:
            if value:
                return variables.BooleanValue(
                    self.TestPoint(
                        value.value,
                        None,
                        None))
            else:
                return variables.BooleanValue()
        elif value.Cardinality() in (variables.Cardinality.ordered,
                                     variables.Cardinality.multiple):
            if value:
                if value.Cardinality() == variables.Cardinality.multiple:
                    vlist = list(dict_keys(value.value))
                else:
                    vlist = value.value
                for value in vlist:
                    if self.TestPoint(value, None, None):
                        return variables.BooleanValue(True)
                return variables.BooleanValue(False)
            else:
                return variables.BooleanValue()
        elif value.Cardinality() is None:
            return variables.BooleanValue()
        else:
            raise core.ProcessingError("Record values not allowed in Inside")


class LT(NOperator):

    """The lt operator takes two sub-expressions which must both have single
    cardinality and have a numerical base-type. The result is a single boolean
    with a value of true if the first expression is numerically less than the
    second and false if it is greater than or equal to the second::

            <xsd:group name="lt.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'lt')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "lt requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            # grab two floats
            v1v = float(v1.value)
            v2v = float(v2.value)
            return variables.BooleanValue(v1v < v2v)
        else:
            return variables.BooleanValue()


class GT(NOperator):

    """The gt operator takes two sub-expressions which must both have single
    cardinality and have a numerical base-type. The result is a single boolean
    with a value of true if the first expression is numerically greater than
    the second and false if it is less than or equal to the second::

            <xsd:group name="gt.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'gt')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "gt requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            # grab two floats
            v1v = float(v1.value)
            v2v = float(v2.value)
            return variables.BooleanValue(v1v > v2v)
        else:
            return variables.BooleanValue()


class LTE(NOperator):

    """The lte operator takes two sub-expressions which must both have single
    cardinality and have a numerical base-type. The result is a single boolean
    with a value of true if the first expression is numerically less than or
    equal to the second and false if it is greater than the second::

            <xsd:group name="lte.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'lte')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "lte requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            # grab two floats
            v1v = float(v1.value)
            v2v = float(v2.value)
            return variables.BooleanValue(v1v <= v2v)
        else:
            return variables.BooleanValue()


class GTE(NOperator):

    """The gte operator takes two sub-expressions which must both have single
    cardinality and have a numerical base-type. The result is a single boolean
    with a value of true if the first expression is numerically less than or
    equal to the second and false if it is greater than the second::

            <xsd:group name="durationGTE.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'gte')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "gte requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            # grab two floats
            v1v = float(v1.value)
            v2v = float(v2.value)
            return variables.BooleanValue(v1v >= v2v)
        else:
            return variables.BooleanValue()


class DurationLT(NOperator):

    """The durationLT operator takes two sub-expressions which must both have
    single cardinality and base-type duration. The result is a single boolean
    with a value of true if the first duration is shorter than the second and
    false if it is longer than (or equal) to the second::

            <xsd:group name="durationLT.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'durationLT')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "lt requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_base_types(
            v1.baseType, v2.baseType, variables.BaseType.duration)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            # grab two floats
            v1v = float(v1.value)
            v2v = float(v2.value)
            return variables.BooleanValue(v1v < v2v)
        else:
            return variables.BooleanValue()


class DurationGTE(NOperator):

    """The durationGTE operator takes two sub-expressions which must both have
    single cardinality and base-type duration. The result is a single boolean
    with a value of true if the first duration is longer (or equal, within the
    limits imposed by truncation) than the second and false if it is shorter
    than the second::

            <xsd:group name="durationGTE.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'durationGTE')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "gte requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_base_types(
            v1.baseType, v2.baseType, variables.BaseType.duration)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            # grab two floats
            v1v = float(v1.value)
            v2v = float(v2.value)
            return variables.BooleanValue(v1v >= v2v)
        else:
            return variables.BooleanValue()


class Sum(NOperator):

    """The sum operator takes 1 or more sub-expressions which all have single
    cardinality and have numerical base-types. The result is a single float or,
    if all sub-expressions are of integer type, a single integer that
    corresponds to the sum of the numerical values of the sub-expressions::

            <xsd:group name="sum.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'sum')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) < 1:
            raise core.ProcessingError(
                "sum requires at least one sub-expression, found %i" %
                len(values))
        variables.check_cardinalities(variables.Cardinality.single,
                                      *[x.cardinality() for x in values])
        base_type = variables.check_numerical_types(
            *[x.baseType for x in values])
        if base_type is None:
            # there are no values to add up, we don't know what type we are
            return variables.SingleValue()
        sum = 0
        for v in values:
            if v:
                sum = sum + v.value
            else:
                if base_type == variables.BaseType.float:
                    return variables.FloatValue()
                else:
                    return variables.IntegerValue()
        if base_type == variables.BaseType.float:
            return variables.FloatValue(float(sum))
        else:
            # sum will still be of type integer at this point
            return variables.IntegerValue(sum)


class Product(NOperator):

    """The product operator takes 1 or more sub-expressions which all have
    single cardinality and have numerical base-types. The result is a single
    float or, if all sub-expressions are of integer type, a single integer that
    corresponds to the product of the numerical values of the sub-expressions::

            <xsd:group name="product.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'product')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) < 1:
            raise core.ProcessingError(
                "product requires at least one sub-expression, found %i" %
                len(values))
        variables.check_cardinalities(variables.Cardinality.single,
                                      *[x.cardinality() for x in values])
        base_type = variables.check_numerical_types(
            *[x.baseType for x in values])
        if base_type is None:
            # there are no values to multiply, we don't know what type we are
            return variables.SingleValue()
        product = 1
        for v in values:
            if v:
                product = product * v.value
            else:
                if base_type == variables.BaseType.float:
                    return variables.FloatValue()
                else:
                    return variables.IntegerValue()
        if base_type == variables.BaseType.float:
            return variables.FloatValue(float(product))
        else:
            # sum will still be of type integer at this point
            return variables.IntegerValue(product)


class Subtract(NOperator):

    """The subtract operator takes 2 sub-expressions which all have single
    cardinality and numerical base-types. The result is a single float or, if
    both sub-expressions are of integer type, a single integer that corresponds
    to the first value minus the second::

            <xsd:group name="subtract.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'subtract')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "subtract requires two sub-expressions, found %i" %
                len(values))
        v1, v2 = values
        base_type = variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if base_type is None:
            # there are no values to subtract, we don't know what type we are
            return variables.SingleValue()
        if v1 and v2:
            # grab two floats
            if base_type == variables.BaseType.float:
                v1v = float(v1.value)
                v2v = float(v2.value)
                return variables.FloatValue(v1v - v2v)
            else:
                return variables.IntegerValue(v1.value - v2.value)
        else:
            if base_type == variables.BaseType.float:
                return variables.FloatValue()
            else:
                return variables.IntegerValue()


class Divide(NOperator):

    """The divide operator takes 2 sub-expressions which both have single
    cardinality and numerical base-types. The result is a single float that
    corresponds to the first expression divided by the second expression::

            <xsd:group name="divide.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'divide')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "divide requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            v1v = float(v1.value)
            v2v = float(v2.value)
            try:
                return variables.FloatValue(v1v / v2v)
            except ZeroDivisionError:
                # if [the second value] is zero or the resulting value is
                # outside the value set defined by float (not including
                # positive and negative infinity) then the operator should
                # result in NULL
                return variables.FloatValue()
        else:
            return variables.FloatValue()


class Power(NOperator):

    """The power operator takes 2 sub-expression which both have single
    cardinality and numerical base-types. The result is a single float that
    corresponds to the first expression raised to the power of the second::

            <xsd:group name="power.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'power')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "power requires two sub-expressions, found %i" % len(values))
        v1, v2 = values
        variables.check_numerical_types(v1.baseType, v2.baseType)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            v1v = float(v1.value)
            v2v = float(v2.value)
            try:
                return variables.FloatValue(math.pow(v1v, v2v))
            except ValueError:
                # If the resulting value is outside the value set defined by
                # float (not including positive and negative infinity) then the
                # operator shall result in NULL
                return variables.FloatValue()
        else:
            return variables.FloatValue()


class IntegerDivide(NOperator):

    """The integer divide operator takes 2 sub-expressions which both have
    single cardinality and base-type integer. The result is the single integer
    that corresponds to the first expression (x) divided by the second
    expression (y) rounded down to the greatest integer (i) such that
    i<=(x/y)::

            <xsd:group name="integerDivide.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'integerDivide')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "integerDivide requires two sub-expressions, found %i" %
                len(values))
        v1, v2 = values
        variables.check_base_types(
            v1.baseType, v2.baseType, variables.BaseType.integer)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            try:
                return variables.IntegerValue(v1.value // v2.value)
            except ZeroDivisionError:
                return variables.IntegerValue()
        else:
            return variables.IntegerValue()


class IntegerModulus(NOperator):

    """The integer modulus operator takes 2 sub-expressions which both have
    single cardinality and base-type integer. The result is the single integer
    that corresponds to the remainder when the first expression (x) is divided
    by the second expression (y). If z is the result of the corresponding
    integerDivide operator then the result is x-z*y::

            <xsd:group name="integerModulus.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="2" maxOccurs="2"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'integerModulus')

    def evaluate(self, state):
        values = list(self.evaluate_children(state))
        if len(values) != 2:
            raise core.ProcessingError(
                "integerModulus requires two sub-expressions, found %i" %
                len(values))
        v1, v2 = values
        variables.check_base_types(
            v1.baseType, v2.baseType, variables.BaseType.integer)
        variables.check_cardinalities(
            v1.Cardinality(), v2.Cardinality(), variables.Cardinality.single)
        if v1 and v2:
            try:
                return variables.IntegerValue(v1.value % v2.value)
            except ZeroDivisionError:
                return variables.IntegerValue()
        else:
            return variables.IntegerValue()


class Truncate(UnaryOperator):

    """The truncate operator takes a single sub-expression which must have
    single cardinality and base-type float. The result is a value of base-type
    integer formed by truncating the value of the sub-expression towards zero::

            <xsd:group name="truncate.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'truncate')

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.float)
        variables.check_cardinalities(
            value.Cardinality(), variables.Cardinality.single)
        if value:
            return variables.IntegerValue(int(value.value))
        else:
            return variables.IntegerValue()


class Round(UnaryOperator):

    """The round operator takes a single sub-expression which must have single
    cardinality and base-type float. The result is a value of base-type integer
    formed by rounding the value of the sub-expression. The result is the
    integer n for all input values in the range [n-0.5,n+0.5)::

            <xsd:group name="round.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'round')

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.float)
        variables.check_cardinalities(
            value.Cardinality(), variables.Cardinality.single)
        if value:
            return variables.IntegerValue(int(math.floor(value.value + 0.5)))
        else:
            return variables.IntegerValue()


class IntegerToFloat(UnaryOperator):

    """The integer to float conversion operator takes a single sub-expression
    which must have single cardinality and base-type integer. The result is a
    value of base type float with the same numeric value::

            <xsd:group name="integerToFloat.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="1" maxOccurs="1"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'integerToFloat')

    def evaluate(self, state):
        value = self.Expression.evaluate(state)
        variables.check_base_types(value.baseType, variables.BaseType.integer)
        variables.check_cardinalities(
            value.Cardinality(), variables.Cardinality.single)
        if value:
            return variables.FloatValue(float(value.value))
        else:
            return variables.FloatValue()


class CustomOperator(NOperator):

    """The custom operator provides an extension mechanism for defining
    operations not currently supported by this specification::

            <xsd:attributeGroup name="customOperator.AttrGroup">
                    <xsd:attribute name="class" type="identifier.Type"
                    use="optional"/>
                    <xsd:attribute name="definition" type="uri.Type"
                    use="optional"/>
                    <xsd:anyAttribute namespace="##other"/>
            </xsd:attributeGroup>

            <xsd:group name="customOperator.ContentGroup">
                    <xsd:sequence>
                            <xsd:group ref="expression.ElementGroup"
                            minOccurs="0" maxOccurs="unbounded"/>
                    </xsd:sequence>
            </xsd:group>"""
    XMLNAME = (core.IMSQTI_NAMESPACE, 'customOperator')
    XMLATTR_class = "customClass"
    XMLATTR_definition = 'definition'

    def __init__(self, parent):
        NOperator.__init__(self, parent)
        self.customClass = None
        self.definition = None

    def evaluate(self, state):
        raise core.ProcessingError(
            "customOperator.%s not supported" %
            (self.customClass
                if self.customClass is not None
                else "<unknown>"))
